//
//  machine.h
//	Yon
//
//	Created by Miso on 13.8.2009.
//	Copyright 2009 Miso Antonic. All rights reserved.
//
//

#ifndef MACHINE_H
#define MACHINE_H

#include "common.h"
#include "string.h"
#include "object.h"
#include "compiler_types.h"
#include "compiler.h"

extern bool haltSignal;

#define MAX_CALL_POOL_SLOT_SIZE		20
#define MAX_CALL_POOL_STACK_SIZE	20

#define OBJECTREFS_IN_SLOT  (sizeof(Slot) / sizeof(ObjectRef))

// This is in sizeof(ObjectRef) units. Slots take exactly 2x the sizeof(ObjectRef)
#define MAX_CALL_POOL_DATA_SIZE     (MAX_CALL_POOL_SLOT_SIZE * OBJECTREFS_IN_SLOT) + MAX_CALL_POOL_STACK_SIZE

#define BOOL(object)            ((BoolObjectRef)object)
#define REAL(object)            ((RealObjectRef)object)
#define FUNCTION(object)        ((FunctionObjectRef)object)
#define NATIVE_FUNCTION(object) ((NativeFunctionObjectRef)object)
#define CALL(object)            ((CallObjectRef)object)
#define STRING(object)          ((StringObjectRef)object)
#define ARRAY(object)           ((ArrayObjectRef)object)
#define DATA(object)            ((DataObjectRef)object)

struct Machine
{
	BYTE* bytecodes;
	
	BytecodeInfo* bytecodeInfos;
	ULONG bytecodeInfoCount;
		
	CHAR** symbolTable;
	ULONG symbolTableLength;
    
    CHAR** sourceIdTable;
    ULONG sourceIdTableLength;
	
	ObjectRef objects; // linked list
    ObjectRef cache;
	
	ObjectRef pool[numberOfObjectTypes];  // objects (except contexts) are stored here for recycling
	CallObjectRef callPools[MAX_CALL_POOL_DATA_SIZE]; // contexts of size [0..(MAX_POOL_CONTEXT_SLOT_SIZE*OBJECTREFS_IN_SLOT + MAX_POOL_CONTEXT_STACK_SIZE-1)] are stored here for recycling
		
	SourcePosition position;
	ErrorRef error;
	wstring internalErrorMessage;
	
	ObjectRef objectNil;
	ObjectRef objectYes;
	ObjectRef objectNo;
	
	ObjectRef primitivePrototypes[numberOfObjectTypes];
    
    ObjectRef objectInternalError;
    
    ObjectRef objectResult;
    ObjectRef objectParameter;
    // REMARK: these have to be roots in GC
    
    BYTE* bytecodeData;

    ~Machine();
};
typedef Machine* MachineRef;


MachineRef machineNew(const char* path);
MachineRef machineNew(BYTE* bytes, ULONG byteCount);
MachineRef machineNew(CompilerRef compiler);

ObjectRef machineRun(MachineRef machine, ObjectRef parameter);

ObjectRef newObjectFromPool(ObjectRef* objects, ObjectRef* pool, BYTE type, ULONG size);
StringObjectRef newStringObject(ObjectRef* objects, ObjectRef* pool, const CHAR* data, ULONG length);
StringObjectRef newStringObject(MachineRef machine, const CHAR* data, ULONG length);
StringObjectRef newStringObjectDontCopy(ObjectRef* objects, ObjectRef* pool, CHAR* data, ULONG length);
RealObjectRef newRealObject(ObjectRef* objects, ObjectRef* pool, DOUBLE value);
ArrayObjectRef newArrayObject(ObjectRef* objects, ObjectRef* pool, ObjectRef* data, ULONG length);
ArrayObjectRef newArrayObject(MachineRef machine, ObjectRef* data, ULONG length);
DataObjectRef newDataObject(ObjectRef* objects, ObjectRef* pool, BYTE* bytes, ULONG size);

void printObject(MachineRef machine, ObjectRef object);
ULONG symbolForCString(CHAR** symbolTable, ULONG symbolTableLength, const CHAR* s);

#endif /* MACHINE_H */