//
//  object.h
//  Yon
//
//  Created by Miso on 13.8.2009.
//  Copyright 2009 Miso Antonic. All rights reserved.
//
//

#ifndef OBJECT_H
#define OBJECT_H

#include "common.h"
//#include "assert.h"
#include "language.h"
#include <iostream>
#include "string.h"

typedef enum
{
	typeReal = 0,
    typeCall = 1,
    typeBigCall = 2,
    typeNil = 3,
    typeFunction = 4,
    typeBool = 5,
    typeNativeFunction = 7,
    typeObject = 8,
    typeString = 9,
    typeArray = 10,
    _typeData = 13,
    typeProxy = 15,
    numberOfObjectTypes
} ObjectType;

#define TAG_VISITED					1
#define TAG_CACHED					2
#define TAG_GARBAGE                 4
// debugging
#define TAG_DEBUG					8

struct Object;

typedef struct Slot
{
	ULONG symbol; // sizeof(ULONG) must equal sizeof(Object*) because otherwise call pooling will not work
	Object* value;
} Slot;

struct Machine;
typedef Machine* MachineRef;

typedef Object* (*NativeFunctionRef)(Object* self, Object** parameters, MachineRef machine);

#define OBJECT_FIELDS \
    BYTE tag; \
    BYTE type; \
    \
    Object* next; \
    Object* prev; \
    \
    Slot* slots; \
    ULONG slotCount; // TODO: smaller var type


typedef struct Object
{
    OBJECT_FIELDS;
} Object;
typedef Object* ObjectRef;


typedef struct RealObject
{
    OBJECT_FIELDS;
    DOUBLE value;
} RealObject;
typedef RealObject* RealObjectRef;

typedef struct StringObject
{
    OBJECT_FIELDS;
    ULONG size; 
    CHAR* data;
} StringObject;
typedef StringObject* StringObjectRef;

typedef struct ArrayObject
{
    OBJECT_FIELDS;
    ULONG size;
    ObjectRef* data;
} ArrayObject;
typedef ArrayObject* ArrayObjectRef;

struct CallObject;

typedef struct FunctionObject
{
    OBJECT_FIELDS;
    BYTE numberOfParameters; 
    BYTE* startAddress; 
    ULONG numberOfSlotsNeededForCall; 
    ULONG stackSizeNeeded; 
    ULONG* parameters; 
    CallObject* parentCall; 
    CallObject* frozenCall;
} FunctionObject;
typedef FunctionObject* FunctionObjectRef;

typedef struct NativeFunctionObject
{
    OBJECT_FIELDS;
    BYTE numberOfParameters; 
    NativeFunctionRef address; 
    ObjectRef curriedSelf;
} NativeFunctionObject;
typedef NativeFunctionObject* NativeFunctionObjectRef;

typedef struct CallObject
{
    OBJECT_FIELDS;
    ULONG slotCapacity;
    ULONG stackCapacity;
    
    CallObject* parentCall;
    CallObject* parent;   // lexical parent
    ObjectRef* stack;
    ObjectRef* stackTop; // this contains a valid value if this call is not active. If it is, stackTop variable in run function holds the valid value instead.
    CallObject* yieldedFromCall; // when using yield-from, series of calls can be frozen. If this the topmost call, yieldedFromCall contains the "from" call (point of cut of spagetti stack).
    BYTE* bytecodeAddress;
    BYTE* errorBytecode;
    ULONG findinPrimitive;
    
    FunctionObjectRef calledFunction;
    
    // This object's size is dynamic, depending on slot and stack capacity
    // Slot* slotsHere
    // ObjectRef stackHere
} CallObject;
typedef CallObject* CallObjectRef;

typedef struct BoolObject
{
    OBJECT_FIELDS;
    bool value;
} BoolObject;
typedef BoolObject* BoolObjectRef;

typedef struct DataObject
{
    OBJECT_FIELDS;
    BYTE* bytes;
	ULONG size;
} DataObject;
typedef DataObject* DataObjectRef;

static inline void objSetSlot(ObjectRef object, ULONG symbol, ObjectRef value)
{
	for (int i = 0; i < object->slotCount; i++) {
		if (object->slots[i].symbol == symbol)
		{
			object->slots[i].value = value;
			return;
		}
	}
	object->slotCount++;
	REALLOC_ARRAY(object->slots, Slot, object->slotCount);
	object->slots[object->slotCount-1].symbol = symbol;
	object->slots[object->slotCount-1].value = value;
}

static inline ObjectRef objGetSlot(ObjectRef object, ULONG symbol) {
	for (int i = 0; i < object->slotCount; i++) {
		if (object->slots[i].symbol == symbol)
			return object->slots[i].value;
	}
	return NULL;
}

// For adding non-var and non-argument slots to context, use addSlot macro instead. since sufficient capacity is not guaranteed
//#define addSlotToCall(object, s, v) 

static inline void callAddSlot(CallObjectRef object, ULONG symbol, ObjectRef value)
{
	for (int i = 0; i < object->slotCount; i++) {
		if (object->slots[i].symbol == symbol) {
			object->slots[i].value = value;
			return;
		}
	}
	ULONG count = ++object->slotCount;
//	if (count > object->slotCapacity)
//		assert(count <= object->slotCapacity);
	object->slots[count-1].symbol = symbol;
	object->slots[count-1].value = value;
} 

static inline bool callSetSlot(CallObjectRef call, ULONG symbol, ObjectRef value)
{
	while (call) {
		for (int i = 0; i < call->slotCount; i++) {
			if (call->slots[i].symbol == symbol) {
				call->slots[i].value = value;
				return true;
			}
		}
		call = call->parent;
	}
	return false;
}

static inline ObjectRef callGetSlot(CallObjectRef call, ULONG symbol)
{
	while (call) {
		ObjectRef result = objGetSlot((ObjectRef)call, symbol);
		if (result) return result;
		call = call->parent;
	}
	return NULL;
}

#endif /* OBJECT_H */